/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */


#ifndef CLEAN__
#define CLEAN__

#include <iostream>
#include <set>
#include <vector> 
#include <iterator>
#include <string.h>
#include <algorithm>
#include <fstream>

#include "globalVars.h"
#include "GeneComp.h"
#include "seq.h"

using namespace std;

/****************************************************************/
/*********************** CleanFile Class ************************/
/****************************************************************/

template <class GENOME_TYPE>
class CleanFile{
	
	private:
		
		/**
	 	 * fileN variable holds the name of the file (*.anchor) containing the KO's.
		 */
		string fileN;
		
		/**
		 * A set of all legal entries (SEQ<GENOME_TYPE> objects) of the given file.
		 */
		set< SEQ<GENOME_TYPE> >* segments;

		/**
		 * An auxiliary declaration of a set iterator.
		 */ 
		typedef typename std::set< SEQ<GENOME_TYPE> >::iterator itr;
		
		/**
		 * Copies the title of the original file (*.anchor) to the "legal" file (*.anchor.clean).
		 * @param file: FILE* of the *.anchor.
		 * @param legal_file: FILE* of the *.anchor.clean.
		 */ 
		void copy_title(FILE* file,FILE* legal_file)const;
	
		/**
		 * The main method of this class which legalizes the *.anchor file.
		 * @param file: FILE* of the *.anchor file.
		 */
		void legalize_it(FILE* file);
		
		/**
		 * Given one entry from the file *.anchor, the method checks its regularity.
		 * @param buff: The entry itself.
		 * @return 0 if legal, 1 otherwise.
		 */
		int check_it(string buff);
		
		/**
		 * Checks if the given character is a space.
		 * @param c: char
		 * @return bool: true if c is space, false otherwise
		 */
		bool isspace(char c);
		
		/**
		 * Checks if the given character is a digit.
		 * @param c: char
		 * @return bool: true if c is a digit, false otherwise
		 */
		bool isdigit(char c);	

		/**
		 * Skips leading spaces of the given string.
		 * @param ar: string
		 * @return int: The first index of the cell that differs from space
		 */
		int skipspaces(string ar);
	
		/**
		 * Checks if Or is a legal orientation value.
		 * @param Or: string
		 * @return 0 if legal, 1 otherwise.
		 */
		int checkO(string Or);
		
		/**
		 * Checks if Start is a legal number.
		 * @param Start: string
		 * @return 0 if legal, 1 otherwise.
		 */
		int checkS(string Start);
		
		/**
		 * Checks if End is a legal number.
		 * @param End: string
		 * @return 0 if legal, 1 otherwise.
		 */
		int checkE(string End);

		/**
		 * Checks if the KO is legal.
		 * @param KO: string
		 * @return 0 if legal, 1 otherwise.
		 */
		int checkK(string KO);
		

		/**
		 * Converts a string into an unsigned GENOME_TYPE number.
		 * @param str: string
		 * @return The unsigned GENOME_TYPE number.
		 */
		GENOME_TYPE string2GENOMEType(string str);

		/**
		 * Extracts the orientation out of the string Orientation.
		 * @param Orientation: string.
		 * @return + or - according to the extracted value.
		 */
		const char get_or(string Orientation);

		/**
		 * Copies buff to file LG.
		 * @param buff: string.
		 * @param LG: FILE*.
		 */ 
		void copy_it(string buff,FILE* LG)const;

		/**
		 * Splits str into tokens according to the given delimiters.
		 * @param str: string.
		 * @param tokens: vector<string> reference to the place where the result will be saved in.
		 * @param delimiters: string 
		 */
		void Tokenize(const string& str,
                      vector<string>& tokens,
                      const string& delimiters);

	public:
		
		/**
		 * Copies the result of the CleanFile phase into the file fileRes.
		 * @param fileRes: The name of the file where the result will be saved in (*.anchor.clean in our case).
		 */ 
		void getCleanRes(string fileRes);		
		
		/**
		 * Main constructor.
		 * @param filename: The file *.anchor.
		 */
		CleanFile(string& filename);
		
		/**
		 * Destructor.
		 */
		~CleanFile();

		/**
			 * Copies the intersection result between two CleanFile objects into the file fileRes and returns two vector of the intersected entries, one of each CleanFile object.
		 * @param clean: A CleanFile object which will be tested with intersection with the calling object.
			 * @param res1: A vector< SEQ<GENOME_TYPE> > reference where the entries of the calling object (which are a part of the intersection) will be saved in.
		 * @param res2: A vector< SEQ<GENOME_TYPE> > reference where the entries of the "clean" object (which are a part of the intersection) will be saved in.	 
		 * @param fileRes: The name of the file where the intersection result will be saved in.
		 */
void intersection(const CleanFile<GENOME_TYPE>& clean,vector< SEQ<GENOME_TYPE> >& res1,vector< SEQ<GENOME_TYPE> >& res2, string fileRes)const;
		
		/**
		 * The main method of the CleanFile class is MakeLegal which actually invokes the legalizing process.
		 */ 
		void MakeLegal();
};


/************************************************************/

template <class GENOME_TYPE>
void CleanFile<GENOME_TYPE>::getCleanRes(string fileRes)
{
	itr v1,v2;
	FILE* legal_file=fopen(fileRes.c_str(),"w");
	FILE* file=fopen(fileN.c_str(),"r");
	
	if(file==NULL)
	{
		cerr<<"Failed to open file "<<fileN<<endl;
		//errors<<"Failed to open file "<<fileN<<endl;
		//finish.open((input3+string("done")).c_str());

		exit(1);
	}
	if(legal_file==NULL)
	{
		cerr<<"Failed to open file "<<fileRes<<endl;
		//errors<<"Failed to open file "<<fileRes<<endl;
		//finish.open((input3+string("done")).c_str());

		exit(1);
	}
	if(segments==NULL)
	{
		cerr<<"No available results"<<endl;
		//errors<<"No available results"<<endl;
		//finish.open((input3+string("done")).c_str());
		
		exit(1);
	}	

	v1=segments->begin();
	v2=segments->end();

	copy_title(file,legal_file);

	while(v1!=v2)
	{
		copy_it( ((SEQ<GENOME_TYPE>)(*v1)).get_info(),legal_file); 	
		v1++;
	}
	cerr<<"The result can be found in file "<<fileRes<<endl<<endl;
	fclose(file);
	fclose(legal_file);

}

/************************************************************/

template <class GENOME_TYPE>
CleanFile<GENOME_TYPE>::CleanFile(string& filename)
{
		segments=new set< SEQ<GENOME_TYPE> >();		
		fileN=filename;
}	

/************************************************************/

template <class GENOME_TYPE>
CleanFile<GENOME_TYPE>::~CleanFile()
{	
	if ( segments ) delete segments;
}

/************************************************************/

template <class GENOME_TYPE>
void CleanFile<GENOME_TYPE>::MakeLegal()
{	
	FILE* file=fopen(fileN.c_str(),"r");

	if(file==NULL)
	{
		cerr<<"Failed to open file "<<fileN<<endl;
		//errors<<"Failed to open file "<<fileN<<endl;
		//finish.open((input3+string("done")).c_str());

		exit(1);
	}
	
	cerr<<"Legalizing file "<<fileN<<" ..."<<endl;	
		
	legalize_it(file);
	
	fclose(file);
}

/*************************************************************/

template <class GENOME_TYPE>
void CleanFile<GENOME_TYPE>::copy_title(FILE* file,FILE* legal_file)const
{
	char c;
	
	while( (c=fgetc(file)) && (c!=EOF) && (c!='\n') )
		fputc(c,legal_file);

	if(c=='\n')
		fputc(c,legal_file);
}

/**************************************************************/

template <class GENOME_TYPE>
void CleanFile<GENOME_TYPE>::legalize_it(FILE* file)
{	
	char c='a';
	itr s1,s2;
	string buff;

	while(c!=EOF)
	{
		while( (c=fgetc(file)) && (c!='\n') && (c!=EOF) )
			buff+=c;
		if(c==EOF) break;	
		if(buff.length()==0)
		{
			buff="";
			continue;
		}
		check_it(buff);	
		buff="";
	}
	set< SEQ<GENOME_TYPE> >* tmp=new set< SEQ<GENOME_TYPE> >();

	s1=segments->begin();
	s2=segments->end();	
		
	for(;s1!=s2;s1++){
		if(  (*s1).get_dirty() == 0 )
		{
			tmp->insert(*s1);	
		}
	}

	delete segments;
	segments=tmp;
}

/**************************************************************/

template <class GENOME_TYPE>
int CleanFile<GENOME_TYPE>::check_it(string buff)
{
	int flag=0;
	vector<string> tokens;	
	itr v1,v2;
	string tmp=buff,Name,Orientation,Start,End,KO;

	Tokenize(buff,tokens,string("\t"));	
	
	if(tokens.size()!=4)
	{
		return 1;
	}

	Orientation=tokens[0];
	Start=tokens[1];
	End=tokens[2];
	KO=tokens[3];

	if( checkO(Orientation) || checkS(Start) || checkE(End) || checkK(KO) )			
		return 1;
	
	v1=segments->begin();
	v2=segments->end();

	SEQ<GENOME_TYPE>* seq=new SEQ<GENOME_TYPE>();
	KO=KO.substr(skipspaces(KO));
	seq->set_start(string2GENOMEType(Start));
	seq->set_end(string2GENOMEType(End));
	seq->set_KO( KO );
	seq->set_orientation(get_or(Orientation));
	seq->set_info(buff);


	while(v1!=v2)
	{	
		if( (SEQ<GENOME_TYPE>(*v1)).get_KO()==KO )
		{
			flag=1;
			
			if((*v1).get_dirty() == 0)
			{	
				segments->erase(v1);
				seq->set_dirty();
				segments->insert((*seq));
			}
			break;
		}
	
		v1++;
	}

	if(!flag)
		segments->insert( (*seq) );
	else delete seq;

	return 0;

}

/**************************************************************/

template <class GENOME_TYPE>
int CleanFile<GENOME_TYPE>::checkO(string Or)
{	
	int i=0;
	
	i+=skipspaces(Or);
	
	if( Or[i]=='\0')
		return 1;

	if( (Or[i]=='+') || (Or[i]=='-') )
	{	
		i++;
		i+=skipspaces(Or.substr(i));
			
		if( Or[i] == '\0' )
			return 0;
		else
			return 1;
	}
	
	return 0;
}

/*************************************************************/

template <class GENOME_TYPE>
int CleanFile<GENOME_TYPE>::checkS(string Start)
{
	int i=0;
	
	i+=skipspaces(Start);

	if(Start[i]=='\0')
		return 1;
	
	while( isdigit(Start[i]) ) 
		i++;

	i+=skipspaces(Start.substr(i));
	
	if((Start[i]=='\0')||(Start[i]==13)/* for carriage return on windows files format*/)
		return 0;
	
	return 1;
}

/************************************************************/

template <class GENOME_TYPE>
int CleanFile<GENOME_TYPE>::checkE(string End)
{	
	return checkS(End);
}
	
/**********************************************************/

template <class GENOME_TYPE>
int CleanFile<GENOME_TYPE>::checkK(string KO)
{	
	unsigned int i=0,j=0;
	string tmp="K";
	vector<string> tokens;
	
	Tokenize(KO,tokens,",");
	for(;i<tokens.size();i++)
	{	
		j+=skipspaces(tokens[i]);
		if( tokens[i][j]=='\0' )
			return 1;
		if( tokens[i][j]!=tmp[0] )
			return 1;

		if(checkS(tokens[i].substr(j+1)))
			return 1;

		j=0;
	}
	return 0;
}

/**********************************************************/

template <class GENOME_TYPE>
int CleanFile<GENOME_TYPE>::skipspaces(string ar)
{
	int i=0;
	while( isspace(ar[i]) )
		i++;
	return i;
}


/**********************************************************/

template <class GENOME_TYPE>
bool CleanFile<GENOME_TYPE>::isspace(char c)
{
	return ( c==' '); 
}

/**********************************************************/

template <class GENOME_TYPE>
bool CleanFile<GENOME_TYPE>::isdigit(char c)
{
	return ( (c>='0') && (c<='9') ); 
}
	
/**********************************************************/

template <class GENOME_TYPE>
void CleanFile<GENOME_TYPE>::copy_it(string buff,FILE* LG)const
{
	unsigned int i=0;
	
	for(;i<buff.length();i++)
   		fputc(buff[i],LG);
	fputc('\n',LG);
}
/*************************************************************/

template <class GENOME_TYPE>
GENOME_TYPE CleanFile<GENOME_TYPE>::string2GENOMEType(string str)
{
	int i=skipspaces(str);
	GENOME_TYPE num=0;
	
	while( (str[i]!='\0') && (isdigit(str[i])) )
	{
		num*=10;
		num+=str[i]-'0';
		i++;
	}
	
	return num;
}

/*************************************************************/

template <class GENOME_TYPE>
const char CleanFile<GENOME_TYPE>::get_or(string Orientation)
{
	int i=skipspaces(Orientation);
		
	if(Orientation[i]=='+')
		return '+';	
	else
		return '-';
}

/*************************************************************/

template <class GENOME_TYPE>
void CleanFile<GENOME_TYPE>::intersection(const CleanFile<GENOME_TYPE>& clean,vector< SEQ<GENOME_TYPE> >& res1,vector< SEQ<GENOME_TYPE> >& res2, string fileRes)const {
	set< SEQ<GENOME_TYPE> > intersec1,intersec2; 
	itr s1,e1,s2,e2,s3,e3,s4,e4;
	insert_iterator< set< SEQ<GENOME_TYPE> > > iter(intersec1,intersec1.begin());
	vector< SEQ<GENOME_TYPE> >* org1=new vector< SEQ<GENOME_TYPE> >();
	vector< SEQ<GENOME_TYPE> >* org2=new vector< SEQ<GENOME_TYPE> >() ;
	SEQ<GENOME_TYPE> *tmp1,*tmp2;
	GeneComp<GENOME_TYPE> comp;
	tmp1=new SEQ<GENOME_TYPE>();
	tmp2=new SEQ<GENOME_TYPE>();	

	GENOME_TYPE start1,start2,end1,end2;
	string ko1,ko2;
	char or1,or2;

	FILE* resFile;
	int flag=0;
	if(fileRes.length()!=0)
	{
		resFile=fopen(fileRes.c_str(),"w");
		if(resFile==NULL)
		{
			cerr<<"Failed to open file "<<fileRes<<endl;
			//errors<<"Failed to open file "<<fileRes<<endl;
			//finish.open((input3+string("done")).c_str());

			exit(1);
		}
		flag=1;
	}

	if(segments==NULL)
	{
		cerr<<"No available results"<<endl;
		//errors<<"No available results"<<endl;
		//finish.open((input3+string("done")).c_str());
		
		exit(1);
	}	


	s1=segments->begin();
	e1=segments->end();
	s2=clean.segments->begin();
	e2=clean.segments->end();
	
	set_intersection(s1,e1,s2,e2, iter);
	insert_iterator< set< SEQ<GENOME_TYPE> > > iter1(intersec2,intersec2.begin());

	s2=segments->begin();
	e2=segments->end();
	s1=clean.segments->begin();
	e1=clean.segments->end();

	set_intersection(s1,e1,s2,e2, iter1);

	s3=intersec1.begin();
	e3=intersec1.end();
	s4=intersec2.begin();
	e4=intersec2.end();

	if(flag)
	{
		copy_it(string("The intersection result is:"),resFile);
		fputc('\n',resFile);
	}
	
	cerr<<"Finding intersection ..."<<endl;	
	
	while(s3!=e3)
	{
		start1=((SEQ<GENOME_TYPE>)(*(s3))).get_start();	
		start2=((SEQ<GENOME_TYPE>)(*(s4))).get_start();
		end1=((SEQ<GENOME_TYPE>)(*(s3))).get_end();
		end2=((SEQ<GENOME_TYPE>)(*(s4))).get_end();
		ko1=((SEQ<GENOME_TYPE>)(*(s3))).get_KO();
		ko2=((SEQ<GENOME_TYPE>)(*(s4))).get_KO();
		or1=((SEQ<GENOME_TYPE>)(*(s3))).get_orientation();
		or2=((SEQ<GENOME_TYPE>)(*(s4))).get_orientation();

		if(flag)
		{	
			
			copy_it(((SEQ<GENOME_TYPE>)(*(s3))).get_info(),resFile);
			copy_it(((SEQ<GENOME_TYPE>)(*(s4))).get_info(),resFile);
			fputc('\n',resFile);
		}
	
					
		tmp1->set_start(start1);
		tmp2->set_start(start2);
		
		tmp1->set_end(end1);
		tmp2->set_end(end2);

		tmp1->set_orientation(or1);
		tmp2->set_orientation(or2);
		
		tmp1->set_KO(ko1);
		tmp2->set_KO(ko2);

		tmp1->set_info(((SEQ<GENOME_TYPE>)(*(s3++))).get_info());
		tmp2->set_info(((SEQ<GENOME_TYPE>)(*(s4++))).get_info());
		
		org1->insert(org1->begin(),(*tmp1) );	
		org2->insert(org2->begin(),(*tmp2) );
	}

	delete tmp1;	
	delete tmp2;	
	
	sort(org1->begin(),org1->end(),comp);
	sort(org2->begin(),org2->end(),comp);

	res1=(*org1);
	res2=(*org2);

	if(flag)
	{
		fputc('\n',resFile);
	}

	cerr<<"The result can be found in file "<<fileRes<<endl<<endl;

	fclose(resFile);

}	

/*************************************************************/

template <class GENOME_TYPE>
void CleanFile<GENOME_TYPE>::Tokenize(const string& str,vector<string>& tokens,
                      const string& delimiters)
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

/*************************************************************/

#endif
	
