/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */


#ifndef _KIS_
#define _KIS_

#include <iostream>
#include <vector>

#include "seq.h"

/***********************************************************/
/*********************** KIS Class *************************/
/***********************************************************/

template <class GENOME_TYPE>
class KIS{
	
	private:	
		/**
		 * Two vectors of SEQ objects that define both segments of a KIS.
		 */ 
		vector< SEQ<GENOME_TYPE> > *seg1,*seg2;
		
		/**
		 * The orientation of the KIS (+/-).
		 */
		char orientation;
		
		/**
		 * An auxiliary declaration of a vector iterator.
		 */ 
		typedef typename std::vector< SEQ<GENOME_TYPE> >::iterator itr;

	public:
		
		/**
		 * Default constructor.
		 */
		KIS();
		
		/**
		 * Secondary constructor.
		 * @param ORI: The orientation of the KIS.
		 */
		KIS(char ORI);	
		
		/**
		 * Copy constructor.
		 * @param K: KIS<GENOME_TYPE> object reference.
		 */
		KIS(const KIS<GENOME_TYPE>& K);

		/**
		 * Destructor.
		 */
		~KIS();
		
		/**
		 * Prints the coordinates of both segments in the KIS along with their orientation to the given file.
		 * @param whereTo: FILE* 
		 */
		void print(FILE* whereTo);
		
		/**
		 * Assignment operator.
		 * Copies K's fields to the calling object fields.
		 * @param K: const KIS<GENOME_TYPE>&
		 * @return KIS<GENOME_TYPE>&
		 */ 
		KIS<GENOME_TYPE>& operator=(const KIS<GENOME_TYPE>& K);
		
		/**
		 * Adds segment1 to the end of seg1 vector, and segment2 to the end of seg2.
		 * @param segment1: The first desired segment for insertion.
		 * @param segment2: The second desired segment for insertion.
		 */
		void insert_seg(const SEQ<GENOME_TYPE>& segment1,const SEQ<GENOME_TYPE>& segment2);
		
		/**
			 * Combines between the calling object and K by adding both segments of K to the end of both current object's segments, respectively.
		 * @param K: The KIS we want to put together with the calling KIS object.
		 */
		KIS<GENOME_TYPE>& unionKis(const KIS<GENOME_TYPE>& K);	
		
		/**
		 * Returns the orientation of the KIS.
		 * @return char
		 */
		char get_Orientation()const;
		
		/**
		 * Returns the details of both segments held by the calling KIS object.
             * @param start1: A reference to the variable which will save the beginning value of the segment in the first genome.
           	 * @param start2: A reference to the variable which will save the beginning value of the segment in the second genome.		
	             * @param end1: A reference to the variable which will save the ending value of the segment in the first genome.
	     	  * @param end2: A reference to the variable which will save the ending value of the segment in the second genome.
		 */
		void get_Coords(GENOME_TYPE& start1,GENOME_TYPE& start2,GENOME_TYPE& end1,GENOME_TYPE& end2)const; 	
};

/********************************************************/

template <class GENOME_TYPE>
char KIS<GENOME_TYPE>::get_Orientation()const
{
	if(orientation=='+')
		return '+';
	else		
		return '-';
}

/********************************************************/

template <class GENOME_TYPE>
void KIS<GENOME_TYPE>::get_Coords(GENOME_TYPE& start1,GENOME_TYPE& start2,GENOME_TYPE& end1,GENOME_TYPE& end2)const
{
	itr s1,s2,e1,e2;

	s1=seg1->begin();
	s2=seg2->begin();
	e1=seg1->end();
	e2=seg2->end();
	if(orientation=='+')
	{
		start1=(*s1).get_start();
		end1=(*(--e1)).get_end();
		start2=(*s2).get_start();
		end2=(*(--e2)).get_end();
	}
	else
	{
		start1=(*s1).get_start();
		end1=(*(--e1)).get_end();
		start2=(*(--e2)).get_start();
		end2=(*s2).get_end();	
	}
}

/********************************************************/

template <class GENOME_TYPE>
void KIS<GENOME_TYPE>::print(FILE* whereTo)
{
	itr s1,s2,e1,e2;
	
	if(whereTo==NULL)
		return;

	s1=seg1->begin();
	s2=seg2->begin();
	e1=seg1->end();
	e2=seg2->end();
	
	if(orientation=='+')
	{
		if(whereTo)
		{
			fprintf(whereTo,"%ld %ld ",(*s1).get_start(),(*(--e1)).get_end());
			fprintf(whereTo,"%ld %ld +\n",(*s2).get_start(),(*(--e2)).get_end());
		}
		else
		{
			cout<< (*s1).get_start()<<" "<< (*(--e1)).get_end()<<" ";
			cout<< (*s2).get_start()<<" "<< (*(--e2)).get_end()<<" +"<<endl;
		}
	}
	else
	{
		if(whereTo)
		{
			fprintf(whereTo,"%ld %ld ",(*s1).get_start(),(*(--e1)).get_end());
			fprintf(whereTo,"%ld %ld -\n",(*(--e2)).get_start(),(*s2).get_end());
		}
		else
		{
			cout<< (*s1).get_start() <<" "<< (*(--e1)).get_end()<<" ";
			cout<< (*(--e2)).get_start()<<" "<<(*s2).get_end()<<" -"<<endl;
		}
	}
return;
}

/*********************************************************/
template <class GENOME_TYPE>
KIS<GENOME_TYPE>::KIS(char ORI)
{
	seg1=new vector< SEQ<GENOME_TYPE> >();
	seg2=new vector< SEQ<GENOME_TYPE> >();
	orientation=ORI;	
}

/********************************************************/
template <class GENOME_TYPE>
KIS<GENOME_TYPE>::KIS()
{
	seg1=new vector< SEQ<GENOME_TYPE> >();
	seg2=new vector< SEQ<GENOME_TYPE> >();
}

/********************************************************/
template <class GENOME_TYPE>
KIS<GENOME_TYPE>::KIS(const KIS<GENOME_TYPE>& K)
{
	seg1=new vector< SEQ<GENOME_TYPE> >();
	seg2=new vector< SEQ<GENOME_TYPE> >();
	(*this)=K;
}

/********************************************************/

template <class GENOME_TYPE>
KIS<GENOME_TYPE>& KIS<GENOME_TYPE>::operator=(const KIS<GENOME_TYPE>& K)
{
	orientation=K.orientation;
	typedef typename std::vector< SEQ<GENOME_TYPE> >::iterator itr;
	itr v1,v2,v3;
	
	v1=K.seg1->begin();v2=K.seg1->end();
	v3=K.seg2->begin();
	
	while(v1!=v2)
	{
		seg1->push_back(*v1);
		seg2->push_back(*v3);
		v1++;v3++;
	}

	return *this;
}

/********************************************************/

template <class GENOME_TYPE>
KIS<GENOME_TYPE>::~KIS()
{
	delete seg1;
	delete seg2;
}

/********************************************************/
template <class GENOME_TYPE>
void KIS<GENOME_TYPE>::insert_seg(const SEQ<GENOME_TYPE>& segment1,const SEQ<GENOME_TYPE>& segment2)
{

	seg1->push_back(segment1);
	seg2->push_back(segment2);	

	return;	
}

/********************************************************/

template <class GENOME_TYPE>
KIS<GENOME_TYPE>& KIS<GENOME_TYPE>::unionKis(const KIS<GENOME_TYPE>& K)
{
	unsigned int i=0;
	
	for(;i<K.seg1->size();i++)
		this->seg1->push_back( K.seg1[0][i] );

	for(i=0;i<K.seg2->size();i++)
		this->seg2->push_back( K.seg2[0][i] );

	return *this;
}

/********************************************************/

#endif
