/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */


#ifndef _KISES_
#define _KISES_

#include <iostream>
#include <algorithm>
#include <string>

#include "kis.h"
#include "match.h" 
#include "globalVars.h"

using std::string;

/**************************************************************/
/*********************** KISES Class *************************/
/**************************************************************/

template <class GENOME_TYPE>
class KISES{
		
	private:

		/**
		 * A vector of KISs.
		 */ 
		vector< KIS<GENOME_TYPE> > *res;
		
		/**
		 * Binary search for elm in arr and returns the index of the suitable cell.
		 * @param arr: An array of match< SEQ<GENOME_TYPE> > objects (look in match.h).
		 * @param elm: The SEQ<GENOME_TYPE> object we look for.
		 * @param len: The length of arr.
		 * @return int: In case of success, the index of the suitable cell, -1 otherwise.
		 */ 
		int binary_search(const match< SEQ<GENOME_TYPE> >* arr,const SEQ<GENOME_TYPE>& elm,int len)const;
		
		/**
	  	 * This method maps between vector "a" and vector "b" by returning an array of indexes where the i'th cell holds 
		 * the index (j) of the cell in "b" which tells us that the i'th element of "a" is the j'th element of "b". 
		 * @param a: A vector of SEQ<GENOME_TYPE> objects.
		 * @param b: A vector of SEQ<GENOME_TYPE> objects.
		 */
		int* find(const vector< SEQ<GENOME_TYPE> >& a, const vector< SEQ<GENOME_TYPE> >& b)const; 

	public:
		
		/**
		 * Default constructor.
		 */
		KISES();
		
		/**
		 * Destructor.
		 */
		~KISES();
		
		/**
		 * Extract the KISs out of the intersection result (we get it from CleanFile's method) to the file resFile.
		 * @param a: A vector of SEQ<GENOME_TYPE> objects.
		 * @param b: A vector of SEQ<GENOME_TYPE> objects.
		 * @param resFile: The name of the output file.
		 * @return A vector of KIS objects, the KIS table (the same result that was copied to the file).
		 */
vector< KIS<GENOME_TYPE> >& find_KISes(const vector< SEQ<GENOME_TYPE> >& a, const vector< SEQ<GENOME_TYPE> >& b,const char* resFile);
};

/*******************************************************************************/

template <class GENOME_TYPE>
int KISES<GENOME_TYPE>::binary_search(const match< SEQ<GENOME_TYPE> >* arr, const SEQ<GENOME_TYPE>& elm,int len)const
{

	int first=0,last=len-1,mid;
  
	while (first <= last) 
	{
		mid = (first + last) / 2;
		
        	if ( elm > (SEQ<GENOME_TYPE>)(arr[mid].get_elm()) ) 
			first = mid + 1;  
       		else 
			if (elm < (SEQ<GENOME_TYPE>)(arr[mid].get_elm()) ) 
           			last = mid - 1; 
 			else
				return arr[mid].get_match();     
   	}

	return -1;   
}	
	
/*******************************************************************************/

template <class GENOME_TYPE>
KISES<GENOME_TYPE>::KISES()
{
	res=new vector< KIS<GENOME_TYPE> > ();
}

/*******************************************************************************/

template <class GENOME_TYPE>
KISES<GENOME_TYPE>::~KISES()
{
	delete res;
}

/*******************************************************************************/

template <class GENOME_TYPE>
int* KISES<GENOME_TYPE>::find(const vector< SEQ<GENOME_TYPE> >& a, const vector< SEQ<GENOME_TYPE> >& b)const
{

	match< SEQ<GENOME_TYPE> > *arr;
	int len=a.size(),*res=new int[len];
	arr=new match< SEQ<GENOME_TYPE> >[len];

	for(int i=0;i<len;i++ )
	{	
		arr[i].add(b[i]);
		arr[i].set_match(i);
	}

	sort(arr,arr+len);

	for(unsigned int j=0;j<a.size();j++)
		res[j]=binary_search(arr,a[j],len);

	return res;

}
	
/*******************************************************************************/ 

template <class GENOME_TYPE>
vector< KIS<GENOME_TYPE> >& KISES<GENOME_TYPE>:: find_KISes(const vector< SEQ<GENOME_TYPE> >& a, const vector< SEQ<GENOME_TYPE> >& b,const char* resFile)
{
	int n;
	vector< KIS<GENOME_TYPE> > *res=new vector< KIS<GENOME_TYPE> >(),*tmp=new vector< KIS<GENOME_TYPE> >();
	char orientation='-';
	typedef typename std::vector< KIS<GENOME_TYPE> >::iterator itr;
	itr start,end;
	unsigned int i=0,flag,kiscount=0;
	int *des=find(a,b);
	FILE* out=NULL;

	if(resFile)
	{	
		out=fopen(resFile,"w");
		if(out==NULL)
		{
			cerr<<"Failed to open file "<<resFile<<endl;
			//errors<<"Failed to open file "<<resFile<<endl;
			//finish.open((input3+string("done")).c_str());

			exit(1);
		}
	}

	cerr<<"Finding KISes ..."<<endl;

	for(;i<a.size();i++)
	{
		flag=0;
		unsigned int j;
		n=des[i];
	
		if( (a[i].get_orientation())==(b[n].get_orientation()) )
			orientation='+';
		else
			orientation='-';
		int flag1=( (orientation=='+') ? 1:0 );	
		
		if(flag1)
		{
			KIS<GENOME_TYPE> kis('+');
			kis.insert_seg(a[i],b[n]);
			j=n+1;
			i++;
			kiscount++;
		
			while( (i<a.size()) && (j<b.size()) && ((int)(j-1-des[i])==-1))
			{	
			if( a[i].get_orientation()==b[j].get_orientation() )
				{
					kis.insert_seg(a[i++],b[j++]);
					string ko1,ko2; 
					ko1=a[i-1].get_KO();
					ko2=b[j-1].get_KO();
					continue;
				}
				else
				{
					i--;
					flag=1;
					break;
				}
			}
			
			if(!flag) i--;
			
			res->push_back(kis);

		}
		
		else
		{
			KIS<GENOME_TYPE> kis('-');
			kis.insert_seg(a[i],b[n]);
			j=n-1;
			i++;
			kiscount++;
			
			while( (i<a.size()) && (j>=0) && ((j+1-des[i])==1))
			{
			if( a[i].get_orientation()!=b[j].get_orientation() )
				{
					kis.insert_seg(a[i++],b[j--]);
					string ko1,ko2;
					ko1=a[i-1].get_KO();
					ko2=b[j+1].get_KO();
					continue;
				}
				else
				{
					i--;
					flag=1;
					break;
				}
			}

			if(!flag) i--;
		
			res->push_back(kis);
	
		}
		
	}

	if((res->size()>1)&&circular1&&circular2)
	{
		if( (a[0].get_orientation())==(b[des[0]].get_orientation()) )
		{
			if((a[a.size()-1].get_orientation())==(b[des[a.size()-1]].get_orientation())) 	
			{
				(*(--(res->end()))).unionKis(res[0][0]);
				for(i=1;i<res->size();i++)
					tmp->push_back(res[0][i]);
				delete res;
				res=tmp;
			}
		}
		else
		
			if((a[a.size()-1].get_orientation())!=(b[des[a.size()-1]].get_orientation())) 	
			{
				(*(--(res->end()))).unionKis(res[0][0]);
				for(i=1;i<res->size();i++)
					tmp->push_back(res[0][i]);
				delete res;
				res=tmp;
			}
	}

	if(out)
	{	
		unsigned int i=0,j=1;
		for(;i<res->size();i++,j++)
		{
			fprintf(out,"KIS%d ",j);
			(*res)[i].print(out);
		}
	}

	delete[] des;

	cerr<<"The results can be found in file "<<resFile<<endl<<endl;
	
	fclose(out);

	return *res;
}

/*******************************************************************************/ 

#endif

