/*
 *  YASS 1.10
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __THREADS_H_
#define __THREADS_H_
#include "tuple.h"
#include "util.h"
#include "list.h"

/*
 *    Include windows or linux for a thread
 *
 */

#ifdef  BIPROC
#ifdef WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif
#endif



#ifdef CHOOSERBTREE
#include "red_black.h"
#else
#ifdef CHOOSEAVLTREE
#include "avl.h"
#else
#error No tree definition : uncomment either "CHOOSERBTREE" or "CHOOSEAVLTREE" in the "util.h" file
#endif
#endif



/* 
 * forward-backward thread features
 */
typedef struct _FEATURE_ {
  char *chunk_query; 
  int   reverse;     /* on the complementary strain */
  int   left_correction;
  int  *buffer01;    /* buffer for prdyn */
  int   last_point;  /* count the last dot displayed */
  
  tuplelist *first_tl;
  tuplelist *last_tl;

  MA *first_MA;
  MA *last_MA;

  int i_current; /* position on the current chunk */
  int i_chunk;   /* current chunk number */
#ifdef BIPROC
#ifdef WIN32
  HANDLE    thread;
#else
  pthread_t thread;
#endif
#endif

  /* B) Statistics */
#ifdef STATS

  /* a) time */
  double clock_pre;
  double clock_chain;
  double clock_align;
  double clock_post;

  /* b) numbers */
  
  /* b.1) pre-process */
  int    nb_keys_removed;
  /* b.2) assemble */
  int    nb_seeds;
  int    nb_chains_tested;
  int    nb_chains_built;
  /* b.3) align */
  int    nb_ma;
  /* b.4) post-process */
  int    nb_postprocessed_grouping_tests;
  int    nb_postprocessed_ma;
  int    nb_postprocessed_grouping_links;
#endif

  /*	
   * grouping variables
   */
  
  /* The windows size ?? */
  int windows;
  /* The current position (start of the window)*/
  int   win_position;

  /* The global queue of MA, ordered by right pos end*/
  struct _queue_MA Q;
  
  /* We use a tree which can be a red black tree or an avl tree*/
#ifdef CHOOSERBTREE
  struct _rb_tree T;
#else
#ifdef CHOOSEAVLTREE
  struct _avl_tree T;
#else
#error No tree definition : uncomment either "CHOOSERBTREE" or "CHOOSEAVLTREE" inside the "util.h" file
#endif
#endif

} Feature;


extern Feature gv_feature[2];

/*********************
 *
 * Thread definition
 *
 ***********************/

/*
 *  Create Thread
 */

#ifdef  BIPROC
#ifdef WIN32
#define CREATE_THREAD(thread,function,parameters)                                                   \
if ((thread=CreateThread(NULL,0,function, (PVOID) parameters,0, (DWORD *)&gv_thread_result))==NULL){\
      fprintf (stderr,"ERROR:YASS Can not create thread\n");                                        \
      fprintf (stderr,"Use Yass mono-thread");                                                      \
      exit(1);                                                                                      \
   }

#else
#define CREATE_THREAD(thread,function,parameters)                   \
  if (pthread_create(&thread,NULL,function,(void *)parameters)) {   \
      fprintf (stderr,"ERROR:YASS Can not create thread\n");        \
      fprintf (stderr,"Use Yass mono-thread");                      \
      exit(1);                                                      \
   }
#endif
#else
#define CREATE_THREAD(thread,function,parameters)    { function(parameters); }
#endif                                                              

/*
 *  End Thread
 */

#ifdef BIPROC
#ifdef WIN32
#define END_THREAD() {return 0;}
#else
#define END_THREAD() {pthread_exit(NULL);}
#endif
#else
#define END_THREAD() {return NULL;}
#endif

/*
 *  Wait Thread
 */

#ifdef BIPROC
#ifdef WIN32
#define WAIT_THREAD(THREAD)  { WaitForSingleObject(THREAD, INFINITE); }
#else
#define WAIT_THREAD(THREAD)  { pthread_join (THREAD, (void *) &gv_thread_result); }
#endif
#else
#define WAIT_THREAD(THREAD)  ;
#endif

/*
 *  Affichage des grouping
 */

extern int             regroup_begin  ;
extern int             regroup_end    ;
extern int             filtering_begin;
extern int             filtering_end  ;


#ifdef BIPROC
#ifdef WIN32

/* WINDOWS multi proc */
extern HANDLE regroup_mutex_begin;
extern HANDLE regroup_mutex_end  ;
extern HANDLE filtering_mutex_begin;
extern HANDLE filtering_mutex_end;


#define INIT_MUTEX() {                                                 \
 regroup_mutex_begin=CreateMutex(NULL,FALSE,"regroup_mutex_begin");    \
 regroup_mutex_end  =CreateMutex(NULL,FALSE,"regroup_mutex_end");      \
 filtering_mutex_begin=CreateMutex(NULL,FALSE,"filtering_mutex_begin");\
 filtering_mutex_end=CreateMutex(NULL,FALSE,"filtering_mutex_end");    \
}


#define DISPLAY_BEGIN(NAMEMUTEX,NOMVAR,STR) {                          \
  WaitForSingleObject(NAMEMUTEX,INFINITE);                             \
  NOMVAR++;                                                            \
  if(NOMVAR>=gp_reverse) { fprintf(stderr,"%s",STR);fflush(NULL); }    \
  ReleaseMutex(NAMEMUTEX);                                             \
}


#define DISPLAY_END(NAMEMUTEX,NOMVAR)                                  \
  WaitForSingleObject(NAMEMUTEX,INFINITE);                             \
  NOMVAR++;                                                            \
  if(NOMVAR>=gp_reverse) { fprintf(stderr,"finished\n");fflush(NULL);} \
  ReleaseMutex(NAMEMUTEX);                                            

#else
/* UNIX multi proc*/
extern pthread_mutex_t regroup_mutex_begin;
extern pthread_mutex_t regroup_mutex_end  ;
extern pthread_mutex_t filtering_mutex_begin;
extern pthread_mutex_t filtering_mutex_end  ;


#define INIT_MUTEX()  \
 pthread_mutex_init(&regroup_mutex_begin,NULL);    \
 pthread_mutex_init(&regroup_mutex_end,NULL);      \
 pthread_mutex_init(&filtering_mutex_begin,NULL);  \
 pthread_mutex_init(&filtering_mutex_end,NULL);    \


#define DISPLAY_BEGIN(NAMEMUTEX,NOMVAR,STR) {                           \
  pthread_mutex_lock(&NAMEMUTEX);                                       \
  NOMVAR++;                                                             \
  if (NOMVAR>=gp_reverse)  { fprintf(stderr,"%s",STR);fflush(NULL); }   \
  pthread_mutex_unlock(&NAMEMUTEX);                                     \
}

#define DISPLAY_END(NAMEMUTEX,NOMVAR) {                                  \
  pthread_mutex_lock(&NAMEMUTEX);                                        \
  NOMVAR++;                                                              \
  if (NOMVAR>=gp_reverse)  { fprintf(stderr,"finished\n");fflush(NULL);} \
  pthread_mutex_unlock(&NAMEMUTEX);                                      \
}

#endif 
#else
/* Mono Proc*/
extern int display_regroup_mutex_begin;
extern int regroup_mutex_begin;
extern int regroup_mutex_end  ;
extern int filtering_mutex_begin;
extern int filtering_mutex_end  ;


#define INIT_MUTEX()  ;

#define DISPLAY_BEGIN(NAMEMUT,NOMVAR,STRING) {                  \
   NOMVAR = NOMVAR +1;                                          \
   if((NOMVAR)==2){fprintf(stderr,"%s",(STRING));fflush(NULL);} \
}


#define DISPLAY_END(NAMEMUT,NOMVAR) {                          \
   NOMVAR++;                                                   \
   if(NOMVAR==2) {fprintf(stderr,"finished\n");fflush(NULL);}  \
}

#endif

void init_Feature(Feature * feature);

#endif




