/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */



#ifndef __GA__EUR__LA__SH__
#define __GA__EUR__LA__SH__

#include <iostream>
#include <vector> 
#include <iterator>
#include <string>
#include <algorithm> 
#include <fstream>

#include "kis.h"
#include "seq.h"
#include "umr.h"
#include "lightkis.h"
#include "lightkisComp1.h"
#include "lightkisComp2.h"
#include "globalVars.h"

using namespace std;

#ifdef __cplusplus

extern "C"
{
#endif

#include "Galignment.h"
#include "./modified_yass/localAlignment.h"

#ifdef __cplusplus
}
#endif 

/**************************************************************/
/*********************** PREPRO Class *************************/
/**************************************************************/

template <class GENOME_TYPE> 
class PREPRO{
	
	private:
		
		/**
		 * The path for the first *.anchor file.
 		 */
		string fileName1;
		
		/**
		 * The path for the second *.anchor file.
		 */	
		string fileName2;
		
		
		/**
		 * The first genome (extracted from file *.genome).
		 */
		string genome1;
		
		/**
		 * The second genome (extracted from file *.genome).
		 */
		string genome2;

		/**
		 * An auxiliary mathod that inverts and transforms a given sequence (A<=>T , G<=>C).
		 * @param arr: char* A pointer to the array including the sequence.
		 */
		void transformSeq(char* arr)const;

		/**
		 * Ectracts the genome from file "f" into the string "buff". Note that no title will be saved in buff, just nucleotides.		 
		 * @param f: FILE* The pointer to the file with the genome.
		 * @param buff: string& A reference to the string where we want to save the genome in.
		 */
		void getGene(FILE* f, string& buff);
		
		/**
		 * An auxiliary mathod that returns the string name+i+"."+count.
		 * @param i: GENOME_TYPE
		 * @param count: long
		 * @param name: string
		 * @return string
		 */
		string labelIt(GENOME_TYPE i,long count,string name)const;
		
		/**
		 * Checks the lengths of the stitched hits against a predefined threshold (stitchMinLen).
		 * @param res: vector< lightKIS<GENOME_TYPE> >&
			 * Note: The stitched hits are also considered as lightKIS objects so the name of the class doesn't bind it to hold only KIS.
		 */
		void checkStitchedHits(vector< lightKIS<GENOME_TYPE> >& res)const;
		
		/**
		 * Unions two stitched hits according to MAGIC's paper.	
		 * @param elm1: lightKIS<GENOME_TYPE>& 
		 * @param elm2: lightKIS<GENOME_TYPE>&
		 */
		void unionHits(lightKIS<GENOME_TYPE>& elm1,lightKIS<GENOME_TYPE>& elm2)const;	
		
		/**
			 * Performs the action of stitching hits on a given local alignment result (which can also be held in a lightKIS object).
		 * @param res: vector< lightKIS<GENOME_TYPE> >&
		 */
		void stitch(vector< lightKIS<GENOME_TYPE> >& res)const;
		
		/**
			 * Refines the gap table by saving the absolute value of the coordinates of "gaps" (gaps holds the extracted unmatched regions. see umr.h)
		 * in "gaps1" (coordinates of first genome) and "gaps2" (coordinates of second genome).
		 * @param gaps: long**
		 * @param gaps1: long** 
		 * @param gaps2: long**
			 * @param len: long& At the beginning of the method it symbolizes the number of unmatched regions in "gaps", at the end it symbolizes the number of unmatched regions in each one of "gaps1" and "gaps2".
		 * @param seq1: string& The first sequence we are working on.
		 * @param seq2: string& The second sequence we are working on.
		 */  	 
		void refineGapTable(long** gaps,long** gaps1,long** gaps2,long& len,string& seq1,string& seq2)const;

		/**
		 * Refines the i'th KIS of the KIS table according to the given refined gaps for both sequences in the KIS.
		 * @param kisses: vector< KIS<GENOME_TYPE> >&
		 * @param gaps1: long** The result of "refineGapTable" method.
		 * @param gaps2: long** The result of "refineGapTable" method.
		 * @param len: long The number of entries in "gaps1" and "gaps2".	
		 * @param i: unsigned int The KIS we want to refine.
		 * @param newkis: vector< lightKIS<GENOME_TYPE> >& The new refined table (includes just the refined KIS).
		 */ 
			void refineKISTable(vector< KIS<GENOME_TYPE> >& kisses,long** gaps1,long** gaps2,long len,unsigned int i,vector< lightKIS<GENOME_TYPE> >& newkis)const;
		
		/**
		 * Refines the i'th hit of the hits table according to the given refined gaps for both sequences in the hit.
		 * @param hits: vector< lightKIS<GENOME_TYPE> >& 
		 * @param gaps1: long** The result of "refineGapTable" method.
		 * @param gaps2: long** The result of "refineGapTable" method.
		 * @param len: long The number of entries in "gaps1" and "gaps2".	
		 * @param i: unsigned int The hit we want to refine.
			 * @param newhit: vector< lightKIS<GENOME_TYPE> >& The new refined table (adds the refined hit to the end of the table newhit).
		 */ 
			void refineHitTable(vector< lightKIS<GENOME_TYPE> >& hits,long** gaps1,long** gaps2,long len,unsigned int i,vector< lightKIS<GENOME_TYPE> >& newhit)const;


	public:
		
		/**
		 * Constructor.
		 * @param f1: string& The path for the first *.anchor file.
		 * @param f2: string& The path for the second *.anchor file.
		 */
		PREPRO(string& f1,string& f2):fileName1(f1),fileName2(f2),genome1(""),genome2(""){}
		
		/**
		 * Returns one of the genomes in buff (according to whichOne flag, 0 for genome1, 1 for genome2).
		 * @param buff: string&
		 * @param whichOne: int
		 */
		void getGenome(string& buff,int whichOne)const;
		
		/**
			 * Global alignment method (uses MAFFT) (used also to globally align every table that is structured from a vector of KIS objects).
		 * @param kisses vector< KIS<GENOME_TYPE> >& 
		 * @param len: long& Will hold the length of the returned string array.
		 	* @return string* An array of the results where string 2*i and 2*i+1 are the results of the global alignment on KIS i.
		 */
		string* gAlign(vector< KIS<GENOME_TYPE> >& kisses,long& len);

		/**
		 * Local alignment method (includes YASS).
		 * @param chunks: string& A parameter which is parallel to a multifasta file given to YASS.
			 * @param whichOne: int Tells us which genome we'd like to include to the local alignment process (0 for genome1, 1 for genome2).
			 * @return vector< lightKIS<GENOME_TYPE> >& A vector of all YASS's results sorted in an alphabetic order according to each chunk's title (whatever comes after each ">").
		 */
		vector< lightKIS<GENOME_TYPE> >& lAlign(string& chunks,int whichOne)const;
		
		/**
		 * Extracts the gaps out of the global alignment results on the KIS table (uses file umr.h).
			 * @param results: string* A pointer to the array of global alignment results where string 2*i and 2*i+1 are the results of the global alignment on KIS i.
		 * @param kisses: vector< KIS<GENOME_TYPE> >& The KIS table.
		 * @return vector< lightKIS<GENOME_TYPE> >& The new and refined KIS table.
		 */
		vector< lightKIS<GENOME_TYPE> >& extractGaps(string* results,vector< KIS<GENOME_TYPE> >& kisses)const;
		
		/**
		 	* Extracts the breakpoints coordinates out of the KIS table "res" taking into consideration the circularity of both genomes.
			 * The method returns a vector of two vector pointers where the first one includes the breakpoint coordinates of the first genome and the second for the second.
		 * @param res: vector< lightKIS<GENOME_TYPE> >& The new KIS table we get after extracting gaps.
		 * @return vector< vector< lightKIS<GENOME_TYPE> >* >&
		 */
		vector< vector< lightKIS<GENOME_TYPE> >* >& extractBPCoords(vector< lightKIS<GENOME_TYPE> >& res);
		
		/**
			 * Extracts the breakpoints out of the genomes and returns an array of two strings where both strings are a multifasta file alike.
	 	 * First string is all breakpoints of the first genome, Second string is all breakpoints of the second genome. 
			 * Each breakpoint is recognized by its title ( >number of genome (1/2)   start of breakpoint   end of breakpoint) 
		 * @param res: vector< vector< lightKIS<GENOME_TYPE> >* >& The result we get from extractBPCoords.
		 * @return string*
		 */
		string* extractBP(vector< vector< lightKIS<GENOME_TYPE> >* >& res);
		
		/**
		 * Converts the positive number num to a string.
		 * @param num: unsigned long
		 * @param result: string& The result of the converted number.
		 */
		void toString(unsigned long num,string& result)const;
		
		/**
			 * The main stitching method that receives the result of the local alignment and performs the procedure of stitching hits on it.
		 * The stitched hits are returned in a vector of lightKIS objects.
		 * @param res: vector< lightKIS<GENOME_TYPE> >&
		 * @return vector< lightKIS<GENOME_TYPE> >&
		 */
		vector< lightKIS<GENOME_TYPE> >& stitchHits(vector< lightKIS<GENOME_TYPE> >& res)const;	
		
		/**
			 * The main global alignment method for stitched hits (used also to globally align every table that is structured from a vector of lightKIS objects).. 
		 * Receives the stitchedHits method's result and performs MAFFT's global alignment procedure on them.
			 * The result of the global alignment is returned as an array of strings where string 2*i and 2*i+1 are the results of the global alignment on SIS i.
		 * @param stitchedHits: vector< lightKIS<GENOME_TYPE> >&
		 * @return string*
		 */
		string* GAOnStitchedHits(vector< lightKIS<GENOME_TYPE> >& stitchedHits)const;
			
		/**
		 * Extracts the gaps out of the global alignment results on the SIS table (uses file umr.h).
			 * @param results: string* A pointer to the array of global alignment results where string 2*i and 2*i+1 are the results of the global alignment on SIS i.
		 * @param hits: vector< KIS<GENOME_TYPE> >& The SIS table.
		 * @return vector< lightKIS<GENOME_TYPE> >& The new and refined SIS table.
		 */
			vector< lightKIS<GENOME_TYPE> >& extractGapsFromHits(string* results,vector< lightKIS<GENOME_TYPE> >& hits)const;
		
		/**
			 * Simply adds the table "add" to the table "kisTable" ("add" is the table we get from the extractGapsFromHits method).
		 * @param add: vector< lightKIS<GENOME_TYPE> >&
		 * @param kisTable: vector< lightKIS<GENOME_TYPE> >&
		 */
		void addToKISTable(vector< lightKIS<GENOME_TYPE> >& add,vector< lightKIS<GENOME_TYPE> >& kisTable)const;		
		
		/**
		 * Prints the table "vec" to the standard output.
		 * @param vec: vector< lightKIS<GENOME_TYPE> >&
		 */ 
		void print(vector< lightKIS<GENOME_TYPE> >& vec)const;
		
		/**
		 * Prints the table "vec" to the path "fileName".
		 * @param vec: vector< lightKIS<GENOME_TYPE> >&
		 * @param fileName: The file where the result would be.
		 * @return int 0 in case of success, 1 in case we failed to open the file, 2 in case the table is empty.
		 */
		int printToFile(vector< lightKIS<GENOME_TYPE> >& vec,string fileName)const;

};

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::toString(unsigned long num,string& result)const
{
	string res="",buff;
	unsigned long tmp=num;
	int i=0;

	if (num == 0) 
	{
		result=string("0");
		return;
	}

	while (tmp > 0)
   	{
      	long tmp1 = tmp / 10; 
		long tmp2 = tmp%10;
	      buff=char(tmp2 + '0'); 
		res=buff+res;
		tmp = tmp1;
	}	

	while(res[i]=='0') i++;	

	result=res.substr(i,res.length());
     	return;
}

/************************************************************************************/

template <class GENOME_TYPE> 
vector< lightKIS<GENOME_TYPE> >& PREPRO<GENOME_TYPE>::lAlign(string& chunks,int whichOne)const
{
	const char* args[5]={"LA","-d","2","-E",eValue.c_str()};
	string buff,name;
	GENOME_TYPE len,tmp;
	LAresult res,p;
	vector< lightKIS<GENOME_TYPE> >* LAres=new vector< lightKIS<GENOME_TYPE> >;
	lightKIS<GENOME_TYPE>* tmp2;
	lightKISComp1<GENOME_TYPE> comp;
	typedef typename std::vector< lightKIS<GENOME_TYPE> >::iterator itr; 
	itr v1,v2;
	

	getGenome(buff,whichOne);

	if(whichOne==0)	
		len=genome2.length();
	else
		len=genome1.length();

	res=LA(5,args,buff.c_str(),chunks.c_str());

	if(res!=NULL)
	{
		p=res->next;
		while(p)
		{
			tmp=0;
			if(p->orientation=='-')
				tmp2=new lightKIS<GENOME_TYPE>(p->start1,p->end1,p->end2,p->start2,'-');
			else
				tmp2=new lightKIS<GENOME_TYPE>(p->start1,p->end1,p->start2,p->end2,'+');

			name=string(p->name);
			tmp2->nameIt(name);
			GENOME_TYPE s1,e1,s2,e2,tmp=0;
			int k=0;
			tmp2->get_Coords(s1,e1,s2,e2);

			while(name[k]!=' ')k++;
			k++;
			while(name[k]!=' ')
			{
				tmp=tmp*10+name[k]-'0';
				k++;
			}
			if(whichOne==0)
			{
				if(((s2+tmp-1)%len)==0)
					if(((e2+tmp-1)%len)==0)
						tmp2->update(s1,e1,len,len);
					else
						tmp2->update(s1,e1,len,(e2+tmp-1)%len);
				else
					if(((e2+tmp-1)%len)==0)
						tmp2->update(s1,e1,(s2+tmp-1)%len,len);
					else
						tmp2->update(s1,e1,(s2+tmp-1)%len,(e2+tmp-1)%len);
			}

			else
			{	
				if(((s2+tmp-1)%len)==0)
					if(((e2+tmp-1)%len)==0)
						tmp2->update(len,len,s1,e1);
					else
						tmp2->update(len,(e2+tmp-1)%len,s1,e1);
				else
					if(((e2+tmp-1)%len)==0)
						tmp2->update((s2+tmp-1)%len,len,s1,e1);
					else
						tmp2->update((s2+tmp)%len,(e2+tmp)%len,s1,e1);
			}

			LAres->push_back(*tmp2);
			delete tmp2;
			p = p->next;
		}
	}

	v1=LAres->begin();
	v2=LAres->end();

	sort(v1,v2,comp);

	return *LAres;
}

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::getGenome(string& buff,int whichOne)const
{
	if( (whichOne>1)||(whichOne<0) )
	{
		cerr<<"Input Error"<<endl;
		return;
	}
	
	if(whichOne==0)
		buff=genome1;
	else
		buff=genome2;
	
	return;
}

/************************************************************************************/
	
template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::transformSeq(char* arr)const
{
	unsigned long len=strlen(arr),i=len,j=0;
	char* tmp=new char[len+1];

	for(;j<len;j++)
		tmp[j]=arr[--i];

	for(j=0;j<len;j++)
	{
		switch(tmp[j])
		{
			case 'G':
					tmp[j]='C';
					break;
			case 'T':
					tmp[j]='A';
					break;
			case 'C':
					tmp[j]='G';
					break;
			case 'A':		
					tmp[j]='T';
					break;
			case 'g':
					tmp[j]='c';
					break;
			case 't':
					tmp[j]='a';
					break;
			case 'c':
					tmp[j]='g';
					break;
			case 'a':		
					tmp[j]='t';
					break;
		}
	}
	tmp[j]='\0';
	j=0;
	while(tmp[j])
		arr[j]=tmp[j++];
	arr[j]=tmp[j];
	delete[] tmp;
}

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::getGene(FILE* f, string& buff)
{
	char c;

	while( (c=fgetc(f)) && (c!=EOF) )
		if(c=='>')
			while( (c=fgetc(f)) && (c!='\n') );
		else
			if( (c!='\n') && (c!=' ') && (c!=13) )
				buff+=c;
}

/************************************************************************************/

template <class GENOME_TYPE> 
string*  PREPRO<GENOME_TYPE>::gAlign(vector< KIS<GENOME_TYPE> >& kisses,long& len)
{
	FILE* file1,*file2;

	string* results=new string[(kisses.size())*2]();
	const unsigned long siz(kisses.size());
	char** tmpp=new char*[2*siz];

	char* tmp[2],*tmp1,*tmp2;
	unsigned long i=0,max=0,indx=siz+1;
	long j=0;
	GENOME_TYPE s1,s2,e1,e2;

	len = 2*kisses.size();
	file1=fopen(fileName1.c_str(),"r");
	file2=fopen(fileName2.c_str(),"r");
	 
	if( (file1==NULL) || (file2==NULL) )
	{
		cerr<<"Failed to open file."<<endl;
		//errors<<"Failed to open file."<<endl;
		//finish.open((input3+string("done")).c_str());

		exit(1);
	}
	
	getGene(file1,genome1);
	getGene(file2,genome2);
	fclose(file1);
	fclose(file2);

	cerr<<"Performing GlobalAlignment ...\n"<<endl;	

	for(;i<siz;i++)
	{		

		kisses[i].get_Coords(s1,s2,e1,e2);
	
		if(e1>=s1)
		{
			tmp1=new char[e1-s1+2];
			for(j=0;s1<=e1;s1++,j++)
				tmp1[j]=genome1[s1-1];
			tmp1[j]='\0';
		}
		
		else
		{	
			j=0;
			tmp1=new char[genome1.length()-s1+e1+2];
			while((unsigned long)s1<genome1.length())
				tmp1[j++]=genome1[s1++];
			s1=0;
			while(s1<e1)
				tmp1[j++]=genome1[s1++];
			tmp1[j]='\0';
		}			

		if(e2>=s2)
		{
			tmp2=new char[e2-s2+2];

			for(j=0;s2<=e2;s2++,j++)
				tmp2[j]=genome2[s2 -1];
			tmp2[j]='\0';
		}	

		else
		{	
			j=0;
			tmp2=new char[genome2.length()-s2+e2+2];
			while( (unsigned long)s2<genome2.length())
				tmp2[j++]=genome2[s2++];	

			s2=0;
			while(s2<e2)
				tmp2[j++]=genome2[s2++];
			tmp2[j]='\0';
		}

		if(kisses[i].get_Orientation()=='-')
			transformSeq(tmp2);

		if( ( ( strlen(tmp1)>strlen(tmp2) )?strlen(tmp1):strlen(tmp2) ) > max )
		{	
			max=strlen(tmp1)>strlen(tmp2)?strlen(tmp1):strlen(tmp2);
			indx=i;
		}
		
		tmpp[2*i]=tmp1;
		tmpp[2*i+1]=tmp2;

	}

	if(indx!=(siz+1))
	{	
		alignIt(tmpp[2*indx],tmpp[2*indx+1],tmp);	

		results[2*indx]=string(tmp[0]);
		results[2*indx+1]=string(tmp[1]);

		delete[] tmpp[2*indx];
		delete[] tmpp[2*indx+1];
		delete[] tmp[0];
		delete[] tmp[1];
	}

	for(i=0;i<siz;i++)
	{

		if(i==indx) 
			continue;

		alignIt(tmpp[2*i],tmpp[2*i+1],tmp);	

		results[2*i]=string(tmp[0]);
		results[2*i+1]=string(tmp[1]);

		delete[] tmpp[2*i];
		delete[] tmpp[2*i+1];
		delete[] tmp[0];
		delete[] tmp[1];
	}

	delete[] tmpp;

	return results;
}

/************************************************************************************/

template <class GENOME_TYPE> 
vector< lightKIS<GENOME_TYPE> >& PREPRO<GENOME_TYPE>::extractGaps(string* results,vector< KIS<GENOME_TYPE> >& kisses)const
{
	long i=0;
	vector< lightKIS<GENOME_TYPE> >* newKISes = new vector< lightKIS<GENOME_TYPE> >();
	long siz = kisses.size();
	if (results==NULL)
		return *newKISes;

	cerr<<"Extracting Gaps ...\n"<<endl;
	
	for(i=0;i<siz;i++)
	{	
		if((results[2*i].length()==0)&&(results[2*i+1].length()==0))
			continue;

		long** gaps = allocateMtx(results[2*i].length()+results[2*i+1].length(),3);
		long len=extractUnmatchedRegions(results[2*i],results[2*i+1],gaps);


		long** gaps1 = allocateMtx(len,2);
		long** gaps2 = allocateMtx(len,2);
	
		refineGapTable(gaps,gaps1,gaps2,len,results[2*i],results[2*i+1]);
		refineKISTable(kisses,gaps1,gaps2,len,i,*newKISes);		
		
		freeMtx(gaps,len);
		freeMtx(gaps1,len);
		freeMtx(gaps2,len);
	}

	return *newKISes;
}

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::refineGapTable(long** gaps,long** gaps1,long** gaps2,long& len,string& seq1,string& seq2)const
{
	long j=0,i=0,len1=seq1.length(),len2=seq2.length();
	long *tmp1=new long[len1],*tmp2=new long[len2];

	if(seq1[0]=='-')
		tmp1[0]=1;
	else
		tmp1[0]=0;
	for(j=1;j<len1;j++)
		if(seq1[j]=='-')
			tmp1[j]=tmp1[j-1]+1;
		else
			tmp1[j]=tmp1[j-1];

	if(seq2[0]=='-')
		tmp2[0]=1;
	else
		tmp2[0]=0;
	for(j=1;j<len2;j++)
		if(seq2[j]=='-')
			tmp2[j]=tmp2[j-1]+1;
		else
			tmp2[j]=tmp2[j-1];

	for(j=0;j<len;j++)
	{
		if(gaps[j][2]==1)
		{
			
			gaps1[i][0]=gaps[j][0]-tmp1[gaps[j][0]];
			if(gaps1[i][0]<0)
				gaps1[i][0]=0;
			gaps1[i][1]=gaps[j][1]-tmp1[gaps[j][1]];
			if(gaps1[i][1]<0)
         			gaps1[i][1]=0;

			gaps2[i][0]=gaps[j][0]-tmp2[gaps[j][0]];
			if(gaps2[i][0]<0)
         			gaps2[i][0]=0;
			gaps2[i][1]=gaps[j][1]-tmp2[gaps[j][1]];
			if(gaps2[i][1]<0)
        			gaps2[i][1]=0;
			
			i++;
		}
	}

	len = i;

	delete[] tmp1;
	delete[] tmp2;
}

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::refineKISTable(vector< KIS<GENOME_TYPE> >& kisses,long** gaps1,long** gaps2,long len,unsigned int i,vector< lightKIS<GENOME_TYPE> >& newkis)const
{
	int flag=0,notvalid=1;
	GENOME_TYPE s1,s2,e1,e2,s3,e3,s4,e4;
	long j=0,count=1;
	lightKIS<GENOME_TYPE>* tmp;

	kisses[i].get_Coords(s1,s2,e1,e2);	
	if(kisses[i].get_Orientation()=='-')
		flag=1;

	long len1=genome1.length(),len2=genome2.length();
	
	s1%=len1;
	e1%=len1;
	s2%=len2;
	e2%=len2;
		

	for(j=0;j<len;j++)
	{
		if(!flag)
		{
			if(j==0)
			{
				s3=s1;		
				s4=s2;	
				e3=(s1+gaps1[j][0])%len1;
				e4=(s2+gaps2[j][0])%len2;
				if(e3==0) e3=len1;
				if(e4==0) e4=len2;
				
				if((s3==e3)||(s4==e4))
				{
					if(len!=1)
						continue;
					else
						goto X;
				}

				string label=labelIt(i+1,count,string("KIS-"));
				count++;
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'+');
				tmp->nameIt(label);
				newkis.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
			else
			{
				s3=(gaps1[j-1][1]+s1)%len1;
				s4=(gaps2[j-1][1]+s2)%len2;
				e3=(gaps1[j][0]+s1)%len1;
				e4=(gaps2[j][0]+s2)%len2;
				if(e3==0) e3=len1;
				if(e4==0) e4=len2;
				if((s3==e3)||(s4==e4)) continue;
				string label=labelIt(i+1,count,string("KIS-"));
				count++;
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'+');
				tmp->nameIt(label);
				newkis.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
X:
			if( (j==(len-1))&&(e3!=e1)&&(e4!=e2) )
			{
				s3=(gaps1[j][1]+s1)%len1;
				s4=(gaps2[j][1]+s2)%len2;
				e3=e1;
				e4=e2;
				if((s3==e3)||(s4==e4)) continue;
				string label=labelIt(i+1,count,string("KIS-"));
				count++;
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'+');
				tmp->nameIt(label);
				newkis.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
	
		}
		else
		{
			if(j==0)
			{
				s3=s1;
				if(e2<s2)
					s4=(e2+len2-gaps2[j][0])%len2;
				else
					s4=(e2-gaps2[j][0])%len2;

				e3=(gaps1[j][0]+s1)%len1;
				e4=e2;
				if(e3==0) e3=len1;
				
				if((s3==e3)||(s4==e4))
				{
					count--; 
					if(len!=1)
						continue;
					else
						goto Y;
				}

				string label=labelIt(i+1,count,string("KIS-"));
				count++;
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'-');
				tmp->nameIt(label);
				newkis.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
			else
			{
				s3=(gaps1[j-1][1]+s1)%len1;
				if(e2<s2)
					s4=(e2+len2-gaps2[j][0])%len2;
				else
					s4=(e2-gaps2[j][0])%len2;

				e3=(gaps1[j][0]+s1)%len1;
				if(e2<s2)
					e4=(e2+len2-gaps2[j-1][1])%len2;
				else
					e4=(e2-gaps2[j-1][1])%len2;
		
				if(e3==0) e3=len1;
				if(e4==0) e4=len2;
				if((s3==e3)||(s4==e4)) continue;
				string label=labelIt(i+1,count,string("KIS-"));
				count++;
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'-');
				tmp->nameIt(label);
				newkis.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
Y:
			if( (j==(len-1))&&(s4!=s2)&&(e3!=e1) )
			{
				s3=(gaps1[j][1]+s1)%len1;
				s4=s2;
				e3=e1;
				if(e2<s2)
					e4=(e2+len2-gaps2[j][1])%len2;
				else
					e4=(e2-gaps2[j][1])%len2;

				if(e4==0) e4=len2;
				
				if((s3==e3)||(s4==e4)) continue;
				string label=labelIt(i+1,count,string("KIS-"));
				count++;
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'-');
				tmp->nameIt(label);
				newkis.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
		
		}
		
	}

	if(notvalid&&(len==0))
	{	
		string label=labelIt(i+1,count,string("KIS-"));
		if(!flag)
			tmp=new lightKIS<GENOME_TYPE>(s1,e1,s2,e2,'+');
		else
			tmp=new lightKIS<GENOME_TYPE>(s1,e1,s2,e2,'-');
		tmp->nameIt(label);
		newkis.push_back(*tmp);
		delete tmp;
	}
}

/************************************************************************************/

template <class GENOME_TYPE> 
vector< vector< lightKIS<GENOME_TYPE> >* >& PREPRO<GENOME_TYPE>::extractBPCoords(vector< lightKIS<GENOME_TYPE> >& res)
{
	cerr<<"Extracting BreakPoints Coords ... \n"<<endl;

	vector< vector< lightKIS<GENOME_TYPE> >* >* BP = new vector< vector< lightKIS<GENOME_TYPE> >* >;
	BP->push_back(new vector< lightKIS<GENOME_TYPE> >);
	BP->push_back(new vector< lightKIS<GENOME_TYPE> >);
	lightKIS<GENOME_TYPE>* tmp;
	typedef typename std::vector< lightKIS<GENOME_TYPE> >::iterator itr; 
	itr v1,v2;
	GENOME_TYPE s1,s2,e1,e2,s11,e11,s22,e22;	

	if( (res.size()==0) || (res.size()==1) )
	{	
		//delete[] BP;
		return *BP;
	}
	
	res[0].get_Coords(s11,e11,s22,e22);

	for(unsigned long i=1;i<res.size();i++)
	{
		res[i].get_Coords(s1,e1,s2,e2);
		tmp=new lightKIS<GENOME_TYPE>(e11,s1,e22,s2,'+');
		(*BP)[0]->push_back(*tmp);	
		delete tmp;
		s11=s1;e11=e1;s22=s2;e22=e2;
	}
	
	if(circular1==1)
	{
		res[0].get_Coords(s1,e1,s2,e2);
		tmp=new lightKIS<GENOME_TYPE>(e11,s1,e22,s2,'+');
		(*BP)[0]->push_back(*tmp);	
		delete tmp;
	}
	
	v1=res.begin();
	v2=res.end();

	sort(v1,v2);

	res[0].get_Coords(s11,e11,s22,e22);

	for(unsigned long i=1;i<res.size();i++)
	{
		res[i].get_Coords(s1,e1,s2,e2);
		tmp=new lightKIS<GENOME_TYPE>(e11,s1,e22,s2,'+');
		(*BP)[1]->push_back(*tmp);	
		delete tmp;
		s11=s1;e11=e1;s22=s2;e22=e2;
	}
	
	if(circular2==1)
	{
		res[0].get_Coords(s1,e1,s2,e2);
		tmp=new lightKIS<GENOME_TYPE>(e11,s1,e22,s2,'+');
		(*BP)[1]->push_back(*tmp);	
		delete tmp;
	}


	return *BP;
}

/************************************************************************************/

template <class GENOME_TYPE> 
string* PREPRO<GENOME_TYPE>::extractBP(vector< vector< lightKIS<GENOME_TYPE> >* >& res)
{
	cerr<<"Extracting BreakPoints ... \n"<<endl;
	string* chunks=new string[2]();
	string buff;
	GENOME_TYPE len1=genome1.length(),len2=genome2.length();
	
	
	if((res)[0]->size()==0)
		return NULL;
	
	GENOME_TYPE s1,s2,e1,e2;

	for(unsigned long i=0;i<(res)[0]->size();i++)
	{
		(*(res)[0])[i].get_Coords(s1,e1,s2,e2);
		if(s1==e1)
			continue;
		
		chunks[0]+='>';
		this->toString(1,buff);
		chunks[0]+=buff;
		chunks[0]+=' ';
		this->toString(s1,buff);
		chunks[0]+=buff;
		chunks[0]+=' ';
		this->toString(e1,buff);
		chunks[0]+=buff;
		chunks[0]+='\n';
		
			
		if(s1<e1)
		{
			while( (s1<e1)&&(genome1[s1]) )
			{
				chunks[0]+=genome1[s1];
				s1++;
			}
		}
		else
		{
			if(s1>e1)
			{
				while( (s1<len1)&&(genome1[s1]) )
				{
					chunks[0]+=genome1[s1];
					s1++;
				}
				s1=0;
				while((s1<e1)&&(genome1[s1]))
				{
					chunks[0]+=genome1[s1];
					s1++;
				}
			}	
			else
				chunks[0]+=genome1[s1];
		}
		
		chunks[0]+='\n';
	}
	
	for(unsigned long i=0;i<(res)[1]->size();i++)
	{
		(*(res)[1])[i].get_Coords(s1,e1,s2,e2);	
		if(s2==e2)
			continue;

		chunks[1]+='>';
		this->toString(2,buff);
		chunks[1]+=buff;
		chunks[1]+=' ';	
		this->toString(s2,buff);
		chunks[1]+=buff;
		chunks[1]+=' ';
		this->toString(e2,buff);
		chunks[1]+=buff;
		chunks[1]+='\n';
	
		if(s2<e2)
		{
			while( (s2<e2)&&(genome2[s2]) )
			{
				chunks[1]+=genome2[s2];
				s2++;
			}	
		}
		else
		{
			if(s2>e2)
			{
				while( (s2<len2)&&(genome2[s2]) )
				{
					chunks[1]+=genome2[s2];
					s2++;
				}
				s2=0;
				while((s2<e2)&&(genome2[s2]))
				{
					chunks[1]+=genome2[s2];
					s2++;
				}
			}
			else
				chunks[1]+=genome2[s2];
		}
		
		chunks[1]+='\n';	
	}
		
	return chunks;
}

/************************************************************************************/

template <class GENOME_TYPE> 
string PREPRO<GENOME_TYPE>::labelIt(GENOME_TYPE i,long count,string name)const
{
	string label=name,buff;
	this->toString((long)i,buff);
	label+=buff;
	label+='.';
	this->toString(count,buff);
	label+=buff;

	return label;
}

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::stitch(vector< lightKIS<GENOME_TYPE> >& res)const
{
	if((res.size()==0)||(res.size()==1))
		return;
	
	typedef typename std::vector< lightKIS<GENOME_TYPE> >::iterator itr; 
	itr v1,v2;
	lightKISComp2<GENOME_TYPE> comp;
	long len=res.size(),i=0,j=1;
	GENOME_TYPE x,y,a,b,w,z,c,d,x1,y1,a1,b1,w1,z1,c1,d1,l1,l2;
	char or1,or2;
		
	
	v1=res.begin();
	v2=res.end();
	
	sort(v1,v2,comp);
	
	while(i<len)
	{
	
		res[i].get_Coords(x,y,a,b);
		res[j].get_Coords(w,z,c,d);
		x1=x;
		y1=y;
		a1=a;
		b1=b;
		w1=w;
		z1=z;
		c1=c;
		d1=d;
	
		if(j==len)
		{
			i++;
			j=i+1;
			continue;
		}

		or1=res[i].get_Orientation();
		or2=res[j].get_Orientation();
		
		if( or1!=or2 )
		{
			j++;
			continue;
		}
		
		if(circular1)
		{
			if(x>y)
				y1+=genome1.length();
			if(w>z)
				z1+=genome1.length();
		}

		if( w1<y1 )
			l1=0;
		else
			l1=w1-y1;

		if( (unsigned long)l1>stitchDistance )
		{
			i++;
			j=i+1;
		}
		else
		{
			if(circular2)
			{
				if(b<a)
					b1+=genome2.length();
				if(d<c)
					d1+=genome2.length();
			}
			
			if( or1=='+' )
			{
	
				unsigned long dist;
				
				if(b!=b1)
 				{
			        	if((unsigned long)abs(c-b)<stitchDistance)
        			 	{
				         	c1+=genome2.length();
				         	if(d==d1)
				                	d1+=genome2.length();
			         	}
 				}
				else
					if(circular2)
					{	
						dist=genome2.length()-b+c;
						
						if(dist<stitchDistance)
						{
							c1+=genome2.length();
							if(d==d1)
								d1+=genome2.length();
						}
					
					}
								
		
				if( c1>a1 )
				{
					if( c1>b1 )
						l2=c1-b1;
					else
						l2=0;
					if( (unsigned long)abs(long(l1-l2)) > stitchDifference )
					{
						j++;
					}
					else
						if( (unsigned long)l2 > stitchDistance )
						{
							j++;
						}
						else
						{
							unionHits(res[j],res[i]);
							j++;
						}
				}
				else
				{
					j++;
				}
			}	
			if( or1=='-' )
			{
				unsigned long dist;
				
				if(d!=d1)
 				{
			        	if((unsigned long)abs(a-b)<stitchDistance)
        			 	{
				         	a1+=genome2.length();
				         	if(b==b1)
				                	b1+=genome2.length();
			         	}
 				}
				else
					if(circular2)
					{	
						dist=genome2.length()-b+a;
					
						if(dist<stitchDistance)
						{
							a1+=genome2.length();
							if(b==b1)
								b1+=genome2.length();
						}
					
					}

				if( c1<a1 )
				{
					if( a1>d1 )
						l2=a1-d1;
					else
						l2=0;
			
					if( (unsigned long)abs(long(l1-l2)) >stitchDifference )
					{
						j++;
					}
					else
						if( (unsigned long)l2 > stitchDistance )
						{
							j++;
						}
						else
						{
							unionHits(res[j],res[i]);
							j++;
						}
				}
				else
				{
					j++;
				}
			}
		}
	}

}
	
/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::checkStitchedHits(vector< lightKIS<GENOME_TYPE> >& res)const
{
	unsigned long i=0,len=res.size();
		
	for(;i<len;i++)
	{
			GENOME_TYPE  x,y,a,b;
			res[i].get_Coords(x,y,a,b);
			unsigned long len1,len2;
		
			if(y<x)
				len1=genome1.length()-x+y;
			else
				len1=y-x;
			
			if(b<a)
				len2=genome2.length()-a+b;
			else
				len2=b-a;
			
		if( !( ( len1 >stitchMinLen )&&( len2 >stitchMinLen )) )
			res[i].setValid(0);
		else
			res[i].setValid(1); 	
	}
}

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::unionHits(lightKIS<GENOME_TYPE>& elm2,lightKIS<GENOME_TYPE>& elm1)const
{
	GENOME_TYPE x,y,a,b,w,z,c,d;
	
	elm1.get_Coords(x,y,a,b);
	elm2.get_Coords(w,z,c,d);

	y=z;
	if(elm1.get_Orientation()=='+')
		b=d;
	else
		a=c;
	
	elm1.update(x,y,a,b);
}

/************************************************************************************/	

template <class GENOME_TYPE> 
vector< lightKIS<GENOME_TYPE> >& PREPRO<GENOME_TYPE>::stitchHits(vector< lightKIS<GENOME_TYPE> >& res)const
{
	unsigned long i=0,len=res.size();	
	vector< lightKIS<GENOME_TYPE> > tmp1;
	vector< lightKIS<GENOME_TYPE> >* result=new vector< lightKIS<GENOME_TYPE> >;
	string name;


	if((res.size()==0)||(res.size()==1))
		return *result;
	
	while(i<len)
	{
		name=res[i].getName();
		vector< lightKIS<GENOME_TYPE> > tmp;

		while((i<len)&&(name==res[i].getName()))
			tmp.push_back(res[i++]);
		stitch(tmp);
		
		for(unsigned long j=0; j<tmp.size(); j++)
		{
			if(tmp[j].getValid()==1)
				tmp1.push_back(tmp[j]);
		}
	}

	checkStitchedHits(tmp1);
	i=1;
	for(unsigned long j=0; j<tmp1.size(); j++)
		if(tmp1[j].getValid()==1)
		{
			string tmp;
			toString(i++,tmp);
			tmp=string("SIS-")+tmp;
			tmp1[j].nameIt(tmp);
			result->push_back(tmp1[j]);
		}

	return *result;
}

/************************************************************************************/			

template <class GENOME_TYPE> 
string* PREPRO<GENOME_TYPE>::GAOnStitchedHits(vector< lightKIS<GENOME_TYPE> >& stitchedHits)const
{
	string* results=new string[(stitchedHits.size())*2]();
	const int siz(stitchedHits.size());
	char** tmpp=new char*[2*siz];

	char* tmp[2],*tmp1,*tmp2;
	long i=0,j=0,indx=-1;
	unsigned long max=0;
	GENOME_TYPE s1,s2,e1,e2,len1=genome1.length(),len2=genome2.length();

	
	cerr<<"Performing Global Alignment On Stitched Hits ...\n"<<endl;	

	for(;i<siz;i++)
	{		
		stitchedHits[i].get_Coords(s1,e1,s2,e2);

		if(e1>=s1)
		{
			tmp1=new char[e1-s1+2];
			for(j=0;s1<=e1;s1++,j++)
				tmp1[j]=genome1[s1-1];
			tmp1[j]='\0';
		}
		
		else
		{	
			j=0;
			tmp1=new char[len1-s1+e1+2];
			while(s1<len1)
				tmp1[j++]=genome1[s1++];
			s1=0;
			while(s1<e1)
				tmp1[j++]=genome1[s1++];
			tmp1[j]='\0';
		}			

		if(e2>=s2)
		{
			tmp2=new char[e2-s2+2];

			for(j=0;s2<=e2;s2++,j++)
				tmp2[j]=genome2[s2 -1];
			tmp2[j]='\0';
		}	

		else
		{	
			j=0;
			tmp2=new char[len2-s2+e2+2];
			while(s2<len2)
				tmp2[j++]=genome2[s2++];	

			s2=0;
			while(s2<e2)
				tmp2[j++]=genome2[s2++];
			tmp2[j]='\0';
		}

		if(stitchedHits[i].get_Orientation()=='-')
			transformSeq(tmp2);

		if( ( ( strlen(tmp1)>strlen(tmp2) )?strlen(tmp1):strlen(tmp2) ) > max )
		{	
			max=strlen(tmp1)>strlen(tmp2)?strlen(tmp1):strlen(tmp2);
			indx=i;
		}

		tmpp[2*i]=tmp1;
		tmpp[2*i+1]=tmp2;
	}

	for(i=0;i<siz;i++)
	{
		alignIt(tmpp[2*i],tmpp[2*i+1],tmp);
		results[2*i]=string(tmp[0]);
		results[2*i+1]=string(tmp[1]);

		delete[] tmpp[2*i];
		delete[] tmpp[2*i+1];
		delete[] tmp[0];
		delete[] tmp[1];
	}
	
	delete[] tmpp;

	return results;
}

/************************************************************************************/		

template <class GENOME_TYPE> 
vector< lightKIS<GENOME_TYPE> >& PREPRO<GENOME_TYPE>::extractGapsFromHits(string* results,vector< lightKIS<GENOME_TYPE> >& hits)const
{
	long i=0;
	vector< lightKIS<GENOME_TYPE> >* newHits = new vector< lightKIS<GENOME_TYPE> >();
	long siz=hits.size();
	if (results==NULL)
		return *newHits;
	
	for(i=0;i<siz;i++)
	{
		
		if((results[2*i].length()==0)&&(results[2*i+1].length()==0))
			continue;
		
		long** gaps = allocateMtx(results[2*i].length()+results[2*i+1].length(),3);
		long len=extractUnmatchedRegions(results[2*i],results[2*i+1],gaps);
	
		long** gaps1 = allocateMtx(len,2);
		long** gaps2 = allocateMtx(len,2);


		refineGapTable(gaps,gaps1,gaps2,len,results[2*i],results[2*i+1]);
		refineHitTable(hits,gaps1,gaps2,len,i,*newHits);		

		freeMtx(gaps,len);
		freeMtx(gaps1,len);
		freeMtx(gaps2,len);
	}

	return *newHits;
}

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::refineHitTable(vector< lightKIS<GENOME_TYPE> >& hits,long** gaps1,long** gaps2,long len,unsigned int i,vector< lightKIS<GENOME_TYPE> >& newhit)const
{
	int flag=0,notvalid=1;
	GENOME_TYPE s1,s2,e1,e2,s3,s4,e3,e4;
	long j=0,count=1;
	lightKIS<GENOME_TYPE>* tmp;

	hits[i].get_Coords(s1,e1,s2,e2);	
	if(hits[i].get_Orientation()=='-')
		flag=1;

	long len1=genome1.length(),len2=genome2.length();
	
	s1%=len1;
	e1%=len1;
	s2%=len2;
	e2%=len2;
		
	for(j=0;j<len;j++)
	{
		if(!flag)
		{
			if(j==0)
			{
				string label=labelIt(i+1,count,string("SIS-"));
				count++;

				s3=s1;		
				s4=s2;	
				e3=(s1+gaps1[j][0])%len1;
				e4=(s2+gaps2[j][0])%len2;

				if(e3==0) e3=len1;
				if(e4==0) e4=len2;
				if((s3==e3)||(s4==e4))
				{
					count--; 
					if(len!=1)
						continue;
					else
						goto X;
				}
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'+');
				tmp->nameIt(label);
				newhit.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
			else
			{
				string label=labelIt(i+1,count,string("SIS-"));
				count++;

				s3=(gaps1[j-1][1]+s1)%len1;
				s4=(gaps2[j-1][1]+s2)%len2;
				e3=(gaps1[j][0]+s1)%len1;
				e4=(gaps2[j][0]+s2)%len2;
 				
				if(e3==0) e3=len1;
 				if(e4==0) e4=len2;
				if((s3==e3)||(s4==e4)) {count--; continue;}
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'+');
				tmp->nameIt(label);
				newhit.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
X:
			if( (j==len-1)&&(e3!=e1)&&(e4!=e2) )
			{
				string label=labelIt(i+1,count,string("SIS-"));
				count++;

				s3=(gaps1[j][1]+s1)%len1;
				s4=(gaps2[j][1]+s2)%len2;
				e3=e1;
				e4=e2;

				if((s3==e3)||(s4==e4)){count--; continue;}
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'+');
				tmp->nameIt(label);
				newhit.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
	
		}
		else
		{
			if(j==0)
			{
				string label=labelIt(i+1,count,string("SIS-"));
				count++;

				s3=s1;
				if(e2<s2)
					s4=(e2+len2-gaps2[j][0])%len2;
				else
					s4=(e2-gaps2[j][0])%len2;

				e3=(gaps1[j][0]+s1)%len1;
				e4=e2;

				if(e3==0) e3=len1;

				if((s3==e3)||(s4==e4))
				{
					count--; 
					if(len!=1)
						continue;
					else
						goto Y;
				}
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'-');
				tmp->nameIt(label);
				newhit.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
			else
			{
				string label=labelIt(i+1,count,string("SIS-"));
				count++;

				s3=(gaps1[j-1][1]+s1)%len1;

				if(e2<s2)
					s4=(e2+len2-gaps2[j][0])%len2;
				else
					s4=(e2-gaps2[j][0])%len2;
					
				e3=(gaps1[j][0]+s1)%len1;
			
				if(e2<s2)
					e4=(e2+len2-gaps2[j-1][1])%len2;
				else
					e4=(e2-gaps2[j-1][1])%len2;

				if(e3==0) e3=len1;
 				if(e4==0) e4=len2;
				if((s3==e3)||(s4==e4)){count--; continue;}
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'-');
				tmp->nameIt(label);
				newhit.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
Y:
			if( (j==len-1)&&(s4!=s2)&&(e3!=e1) )
			{
				string label=labelIt(i+1,count,string("SIS-"));
				count++;

				s3=(gaps1[j][1]+s1)%len1;
				s4=s2;
				e3=e1;
				if(e2<s2)
					e4=(e2+len2-gaps2[j][1])%len2;
				else
					e4=(e2-gaps2[j][1])%len2;
				
				if(e4==0) e4=len2;
				if((s3==e3)||(s4==e4)){count--; continue;}
				tmp=new lightKIS<GENOME_TYPE>(s3,e3,s4,e4,'-');
				tmp->nameIt(label);
				newhit.push_back(*tmp);
				delete tmp;
				notvalid=0;
			}
		
		}
		
	}

	if(notvalid&&(len==0))
	{	
		string label=labelIt(i+1,count,string("SIS-"));
		if(!flag)
			tmp=new lightKIS<GENOME_TYPE>(s1,e1,s2,e2,'+');
		else
			tmp=new lightKIS<GENOME_TYPE>(s1,e1,s2,e2,'-');
		tmp->nameIt(label);
		newhit.push_back(*tmp);
		delete tmp;
	}
}

/************************************************************************************/

template <class GENOME_TYPE> 
void PREPRO<GENOME_TYPE>::addToKISTable(vector< lightKIS<GENOME_TYPE> >& add,vector< lightKIS<GENOME_TYPE> >& kisTable)const
{
	unsigned long i=0;
	lightKIS<GENOME_TYPE>* tmp;
	lightKISComp2<GENOME_TYPE> comp;
	typedef typename std::vector< lightKIS<GENOME_TYPE> >::iterator itr; 
	itr v1,v2;

	for(;i<add.size();i++)
	{
		tmp=new lightKIS<GENOME_TYPE>(add[i]);
		kisTable.push_back(*tmp);
		delete tmp;
	}
	v1=kisTable.begin();
	v2=kisTable.end();

	sort(v1,v2,comp);

}


/*************************************************************************************/

template <class GENOME_TYPE>
void PREPRO<GENOME_TYPE>::print(vector< lightKIS<GENOME_TYPE> >& vec)const
{
	if(vec.size()==0)	
	{		
		cout<<"Empty!"<<endl;
		return;
	}
					
	cout<<fileName1<<"\t"<<fileName2<<endl;	
	cout<<"START1\t"<<"END1\t"<<"LENGTH1\t"<<"START2\t"<<"END2\t"<<"LENGTH2\t"<<"ORIENTATION\t"<<"IDENTIFIER"<<endl;

	for(unsigned long i=0; i<vec.size(); i++)
	{
		GENOME_TYPE tmp1,tmp2;
		GENOME_TYPE s1,e1,s2,e2;
		vec[i].get_Coords(s1,e1,s2,e2);

		if(s1>e1)
			tmp1=genome1.length()-s1+e1;
		else
			tmp1=e1-s1;
	
		if(s2>e2)
			tmp2=genome2.length()-s2+e2;
		else
			tmp2=e2-s2;
	
		cout<<s1<<"\t"<<e1<<"\t"<<tmp1<<"\t";
		cout<<s2<<"\t"<<e2<<"\t"<<tmp2<<"\t";
		
		if(vec[i].get_Orientation()=='-')
			cout<<"-\t";
		else
			cout<<"+\t";
		cout<<vec[i].getName()<<endl;
	}
}

/************************************************************************************/

template <class GENOME_TYPE>
int PREPRO<GENOME_TYPE>::printToFile(vector< lightKIS<GENOME_TYPE> >& vec,string fileName)const
{
	ofstream whereTo;
	
	whereTo.open(fileName.c_str());
	if(!whereTo)
	{	
		cerr<<"Unable to open file "<<fileName<<endl;
		return 1;
	}
	
	if(vec.size()==0)	
	{		
		whereTo<<"Empty!"<<endl;
		whereTo.close();
		return 2;
	}
					
	whereTo<<fileName1<<"\t"<<fileName2<<endl;
	whereTo<<"START1\t"<<"END1\t"<<"LENGTH1\t"<<"START2\t"<<"END2\t"<<"LENGTH2\t"<<"ORIENTATION\t"<<"IDENTIFIER"<<endl;

	for(unsigned long i=0; i<vec.size(); i++)
	{
		GENOME_TYPE tmp1,tmp2;
		GENOME_TYPE s1,e1,s2,e2;
		vec[i].get_Coords(s1,e1,s2,e2);

		if(s1>e1)
			tmp1=genome1.length()-s1+e1;
		else
			tmp1=e1-s1;
	
		if(s2>e2)
			tmp2=genome2.length()-s2+e2;
		else
			tmp2=e2-s2;
	
		whereTo<<s1<<"\t"<<e1<<"\t"<<tmp1<<"\t";
		whereTo<<s2<<"\t"<<e2<<"\t"<<tmp2<<"\t";
		
		if(vec[i].get_Orientation()=='-')
			whereTo<<"-\t";
		else
			whereTo<<"+\t";
		whereTo<<vec[i].getName()<<endl;
	}
	
	cerr<<"The final PRE-PRO result can be found in file "<<fileName<<endl;

	whereTo.close();
	return 0;
}

/************************************************************************************/


#endif
