/* 
 *  PREPRO 1.0
 *  Created on 1-May-2006
 * 
 *  Copyright (C) 2005-2006 the MAGIC team.
 *  Author(s): Faddy Saad <magicprepro@gmail.com>.
 *
 *
 *  This file is part of MAGIC, an Integrative and Accurate Comparative
 *  Genome Mapping tool.
 *
 *
 *  MAGIC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GCC; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 */



#ifndef _SEQUENCE_
#define _SEQUENCE_

#include <string>
#include <ostream>

using std::string;

/***********************************************************/
/*********************** SEQ Class *************************/
/***********************************************************/

template <class GENOME_TYPE>
class SEQ
{
	private:

		/**
		 * The sequence's coordinates.
		 */
		GENOME_TYPE Start,End;	
		
		/**
		 * The sequence's orientation.
		 */
		char Orientation;
		
		/**
		 * The suitable entry from the file *.anchor.
		 */
		string info;
		
		/**
		 * The sequence's KO.
		 */
		string KO;		
		
		/**
		 * An auxiliary field which was used in cleanfile.h. 
		 * 0 is for valid and 1 is for not valid.
		 */ 
		int dirty;

	public:	
		
		/**
		 * Default constructor.
		 * Initializes the "dirty" field with 0.
		 */ 
		SEQ():dirty(0){}
		
		/**
		 * Copy constructor.
		 * @param G: SEQ<GENOME_TYPE> object reference.
		 */		 
		SEQ(const SEQ<GENOME_TYPE>& G);
		
		/**
		 * Destructor.
		 */
		~SEQ(){}
		
		/**
		 * Assignment operator.
		 * Copies G's fields to the current object fields.
		 * @param G: const SEQ<GENOME_TYPE>&
		 * @return SEQ<GENOME_TYPE>&
		 */
		SEQ<GENOME_TYPE>& operator=(const SEQ<GENOME_TYPE>& G);

		/**	
		 * Sets the KO of the calling object.
		 * @param KO: string.
		 */
		void set_KO(string KO);
		
		/**
		 * Sets the beginning coordinate of the sequence.
		 * @param start: GENOME_TYPE
		 */
		void set_start(GENOME_TYPE start);
		
		/**
		 * Sets the ending coordinate of the sequence.
		 * @param end: GENOME_TYPE
		 */
		void set_end(GENOME_TYPE end);

		/**
		 * Sets the orientation of the sequence.
		 * @param orientation: char
		 */
		void set_orientation(const char orientation);
	
		/**
		 * Sets the "info" field of the sequence.
		 * @param str: string
		 */
		void set_info(string str);

		/**
		 * Sets the "dirty" field of the sequence to 1.
		 */
		void set_dirty();		

		/**
		 * Returns the beginning coordinate of the sequence.
		 * @return GENOME_TYPE
		 */
		GENOME_TYPE get_start()const;
		
		/**
		 * Returns the ending coordinate of the sequence.
		 * @return GENOME_TYPE
		 */
		GENOME_TYPE get_end()const; 
		
		/**
		 * Returns the KO of the sequence.
		 * @return string
		 */
		string get_KO()const;

		/**
		 * Returns the orientation of the sequence.
		 * @return char
		 */
		char get_orientation()const;

		/**
		 * Returns the "info" field of the sequence.
		 * @return string
		 */
		string get_info()const;
		
		/**
		 * Returns the "dirty" field of the sequence.
		 * @return GENOME_TYPE
		 */
		int get_dirty()const;

		/**
		 * Operator < .
		 * Compares between this and g according to KO field. 
		 * @param g: const SEQ<GENOME_TYPE>& 
		 * @return bool
		 */
		bool operator<(const SEQ<GENOME_TYPE>& g)const;
		
		/**
		 * Operator > .
		 * Compares between this and g according to KO field. 
		 * @param g: const SEQ<GENOME_TYPE>&
		 * @return bool
		 */
		bool operator>(const SEQ<GENOME_TYPE>& g)const;
		
		/**
		 * Operator == .
		 * Compares between this and g according to KO field.
		 * @param g: const SEQ<GENOME_TYPE>&
		 * @return bool
		 */
		bool operator==(const SEQ<GENOME_TYPE>& g)const;
};

/************************************************************/

template <class GENOME_TYPE>
void SEQ<GENOME_TYPE>::set_dirty()
{
	dirty=1;
}

/************************************************************/

template <class GENOME_TYPE>
int SEQ<GENOME_TYPE>::get_dirty()const
{
	return this->dirty;
}

/************************************************************/

template <class GENOME_TYPE>
bool SEQ<GENOME_TYPE>::operator>(const SEQ<GENOME_TYPE>& g)const
{
	return (KO > g.KO);
}

/************************************************************/

template <class GENOME_TYPE>
void SEQ<GENOME_TYPE>::set_start(GENOME_TYPE start)
{
	Start=start;
}

/************************************************************/

template <class GENOME_TYPE>
void SEQ<GENOME_TYPE>::set_end(GENOME_TYPE end)
{
	End=end;
}

/************************************************************/

template <class GENOME_TYPE>
void SEQ<GENOME_TYPE>::set_orientation(const char orientation)
{
	if(orientation=='+')
		Orientation='+';
	else
		Orientation='-';
}

/************************************************************/

template <class GENOME_TYPE>
void SEQ<GENOME_TYPE>::set_KO(string str)
{
	KO=str;
}

/************************************************************/

template <class GENOME_TYPE>
void SEQ<GENOME_TYPE>::set_info(string str)
{
	info=str;
}

/************************************************************/

template <class GENOME_TYPE>
GENOME_TYPE SEQ<GENOME_TYPE>::get_start()const
{
	return Start;
}

/************************************************************/		

template <class GENOME_TYPE>
GENOME_TYPE SEQ<GENOME_TYPE>::get_end()const
{
	return End;
}		

/************************************************************/

template <class GENOME_TYPE>
string SEQ<GENOME_TYPE>::get_KO()const
{
	return KO;
}

/************************************************************/
		
template <class GENOME_TYPE>
char SEQ<GENOME_TYPE>::get_orientation()const
{
	if(Orientation=='+')
		return '+';
	return '-';
}		

/************************************************************/

template <class GENOME_TYPE>
string SEQ<GENOME_TYPE>::get_info()const
{
	return info;
}


/************************************************************/

template <class GENOME_TYPE>
bool SEQ<GENOME_TYPE>::operator<(const SEQ<GENOME_TYPE>& g)const
{
	return ( KO < g.KO );
}

/************************************************************/

template <class GENOME_TYPE>
bool SEQ<GENOME_TYPE>::operator==(const SEQ<GENOME_TYPE>& g)const
{	
	return (KO==g.KO);
}

/*************************************************************/

template <class GENOME_TYPE>
SEQ<GENOME_TYPE>& SEQ<GENOME_TYPE>::operator=(const SEQ<GENOME_TYPE>& G)
{
	Start=G.Start;
	End=G.End;
	Orientation=G.Orientation;
	info=G.info;
	KO=G.KO;
	dirty=G.dirty;
	return *this;
}

/*************************************************************/

template <class GENOME_TYPE>
SEQ<GENOME_TYPE>::SEQ(const SEQ<GENOME_TYPE>& G)
{
	Start=G.Start;
	End=G.End;
	Orientation=G.Orientation;
	info=G.info;
	KO=G.KO;
	dirty=G.dirty;
}

/************************************************************/

#endif

